//
//  Mesh.swift
//  PlanetHex
//
//  Created by Don McBrien on 15/03/2020.
//  Copyright © 2020 Don McBrien. All rights reserved.
//

import Foundation
import MetalKit

public typealias float2 = SIMD2<Float>
public typealias float3 = SIMD3<Float>
public typealias float4 = SIMD4<Float>

public struct Vertex {
   let position: float3
   let vertexType: UInt16   // 1 of 3 types (centre, centreLeft or centreRight)
}

public struct Model {
   public let indexTables = [[0, 0, 0, 0, 0, 0].map() { UInt16($0) },
                             [1, 3, 3, 5, 5, 1].map() { UInt16($0) },
                             [2, 2, 4, 4, 6, 6].map() { UInt16($0) }]

   public let localUVCoords = [float2(0.125,0.25),
                               float2(0.005,0.25),
                               float2(0.065,0.496),
                               float2(0.185,0.496),
                               float2(0.245,0.25),
                               float2(0.185,0.004),
                               float2(0.065,0.004)]

   public let uvOffsets = [float2(0.0,0.0),
                           float2(0.0,0.0),
                           float2(0.25,0.0),
                           float2(0.25,0.0),
                           float2(0.5,0.5),
                           float2(0.5,0.5),
                           float2(0.75,0.5),
                           float2(0.75,0.5)]
   
   public let vertices = [Vertex(position: float3(0.8, 0.0, 0.0), vertexType: 0),
                          Vertex(position: float3(0.69282037, 0.0, -0.39999998), vertexType: 1),
                          Vertex(position: float3(0.69282037, 0.0, 0.39999998), vertexType: 2),
                          Vertex(position: float3(0.45923698, 0.23187076, 0.0), vertexType: 0),
                          Vertex(position: float3(0.39771092, 0.23187076, -0.22961847), vertexType: 1),
                          Vertex(position: float3(0.39771092, 0.23187076, 0.22961847), vertexType: 2),
                          Vertex(position: float3(0.30146152, 3.763861e-08, 0.0), vertexType: 0),
                          Vertex(position: float3(0.26107335, 3.763861e-08, -0.15073074), vertexType: 1),
                          Vertex(position: float3(0.26107335, 3.763861e-08, 0.15073074), vertexType: 2),
                          Vertex(position: float3(0.4592369, -0.23187071, 0.0), vertexType: 0),
                          Vertex(position: float3(0.39771083, -0.23187071, -0.22961843), vertexType: 1),
                          Vertex(position: float3(0.39771083, -0.23187071, 0.22961843), vertexType: 2),
                          Vertex(position: float3(0.47071332, 0.22117807, 0.4707133), vertexType: 0),
                          Vertex(position: float3(0.6430063, 0.22117807, 0.17229299), vertexType: 1),
                          Vertex(position: float3(0.17229311, 0.22117807, 0.6430063), vertexType: 2),
                          Vertex(position: float3(0.23689911, 0.12492625, 0.2368991), vertexType: 0),
                          Vertex(position: float3(0.32361016, 0.12492625, 0.08671107), vertexType: 1),
                          Vertex(position: float3(0.08671113, 0.12492625, 0.32361016), vertexType: 2),
                          Vertex(position: float3(0.23689906, -0.12492617, 0.23689905), vertexType: 0),
                          Vertex(position: float3(0.32361013, -0.12492617, 0.08671106), vertexType: 1),
                          Vertex(position: float3(0.086711116, -0.12492617, 0.32361013), vertexType: 2),
                          Vertex(position: float3(0.47071332, -0.22117807, 0.4707133), vertexType: 0),
                          Vertex(position: float3(0.6430063, -0.22117807, 0.17229299), vertexType: 1),
                          Vertex(position: float3(0.17229311, -0.22117807, 0.6430063), vertexType: 2),
                          Vertex(position: float3(6.039832e-08, 0.0, 0.8), vertexType: 0),
                          Vertex(position: float3(0.40000007, 0.0, 0.6928203), vertexType: 1),
                          Vertex(position: float3(-0.3999999, 0.0, 0.6928204), vertexType: 2),
                          Vertex(position: float3(3.4671427e-08, 0.23187076, 0.45923698), vertexType: 0),
                          Vertex(position: float3(0.22961852, 0.23187076, 0.3977109), vertexType: 1),
                          Vertex(position: float3(-0.22961842, 0.23187076, 0.39771095), vertexType: 2),
                          Vertex(position: float3(2.2759712e-08, 3.763861e-08, 0.30146152), vertexType: 0),
                          Vertex(position: float3(0.15073077, 3.763861e-08, 0.26107332), vertexType: 1),
                          Vertex(position: float3(-0.15073071, 3.763861e-08, 0.26107335), vertexType: 2),
                          Vertex(position: float3(3.4671423e-08, -0.23187071, 0.4592369), vertexType: 0),
                          Vertex(position: float3(0.22961847, -0.23187071, 0.3977108), vertexType: 1),
                          Vertex(position: float3(-0.22961837, -0.23187071, 0.39771086), vertexType: 2),
                          Vertex(position: float3(-0.4707133, 0.22117807, 0.4707133), vertexType: 0),
                          Vertex(position: float3(-0.17229302, 0.22117807, 0.6430063), vertexType: 1),
                          Vertex(position: float3(-0.6430063, 0.22117807, 0.17229293), vertexType: 2),
                          Vertex(position: float3(-0.2368991, 0.12492625, 0.2368991), vertexType: 0),
                          Vertex(position: float3(-0.08671109, 0.12492625, 0.32361016), vertexType: 1),
                          Vertex(position: float3(-0.3236102, 0.12492625, 0.08671104), vertexType: 2),
                          Vertex(position: float3(-0.23689905, -0.12492617, 0.23689905), vertexType: 0),
                          Vertex(position: float3(-0.08671107, -0.12492617, 0.32361013), vertexType: 1),
                          Vertex(position: float3(-0.32361013, -0.12492617, 0.08671103), vertexType: 2),
                          Vertex(position: float3(-0.4707133, -0.22117807, 0.4707133), vertexType: 0),
                          Vertex(position: float3(-0.17229302, -0.22117807, 0.6430063), vertexType: 1),
                          Vertex(position: float3(-0.6430063, -0.22117807, 0.17229293), vertexType: 2),
                          Vertex(position: float3(-0.8, 0.0, 1.2079664e-07), vertexType: 0),
                          Vertex(position: float3(-0.69282025, 0.0, 0.40000018), vertexType: 1),
                          Vertex(position: float3(-0.6928203, 0.0, -0.39999998), vertexType: 2),
                          Vertex(position: float3(-0.45923698, 0.23187076, 6.934285e-08), vertexType: 0),
                          Vertex(position: float3(-0.39771083, 0.23187076, 0.2296186), vertexType: 1),
                          Vertex(position: float3(-0.3977109, 0.23187076, -0.22961847), vertexType: 2),
                          Vertex(position: float3(-0.30146152, 3.763861e-08, 4.5519425e-08), vertexType: 0),
                          Vertex(position: float3(-0.2610733, 3.763861e-08, 0.15073083), vertexType: 1),
                          Vertex(position: float3(-0.26107332, 3.763861e-08, -0.15073074), vertexType: 2),
                          Vertex(position: float3(-0.4592369, -0.23187071, 6.9342846e-08), vertexType: 0),
                          Vertex(position: float3(-0.39771074, -0.23187071, 0.22961855), vertexType: 1),
                          Vertex(position: float3(-0.3977108, -0.23187071, -0.22961843), vertexType: 2),
                          Vertex(position: float3(-0.47071332, 0.22117807, -0.47071323), vertexType: 0),
                          Vertex(position: float3(-0.6430063, 0.22117807, -0.17229287), vertexType: 1),
                          Vertex(position: float3(-0.17229328, 0.22117807, -0.6430062), vertexType: 2),
                          Vertex(position: float3(-0.23689911, 0.12492625, -0.23689906), vertexType: 0),
                          Vertex(position: float3(-0.3236102, 0.12492625, -0.08671101), vertexType: 1),
                          Vertex(position: float3(-0.08671121, 0.12492625, -0.32361013), vertexType: 2),
                          Vertex(position: float3(-0.23689906, -0.12492617, -0.23689902), vertexType: 0),
                          Vertex(position: float3(-0.32361013, -0.12492617, -0.086711), vertexType: 1),
                          Vertex(position: float3(-0.0867112, -0.12492617, -0.32361007), vertexType: 2),
                          Vertex(position: float3(-0.47071332, -0.22117807, -0.47071323), vertexType: 0),
                          Vertex(position: float3(-0.6430063, -0.22117807, -0.17229287), vertexType: 1),
                          Vertex(position: float3(-0.17229328, -0.22117807, -0.6430062), vertexType: 2),
                          Vertex(position: float3(9.539905e-09, 0.0, -0.8), vertexType: 0),
                          Vertex(position: float3(-0.40000024, 0.0, -0.6928202), vertexType: 1),
                          Vertex(position: float3(0.3999996, 0.0, -0.69282055), vertexType: 2),
                          Vertex(position: float3(5.476346e-09, 0.23187076, -0.45923698), vertexType: 0),
                          Vertex(position: float3(-0.22961862, 0.23187076, -0.3977108), vertexType: 1),
                          Vertex(position: float3(0.22961825, 0.23187076, -0.397711), vertexType: 2),
                          Vertex(position: float3(3.5948926e-09, 3.763861e-08, -0.30146152), vertexType: 0),
                          Vertex(position: float3(-0.15073085, 3.763861e-08, -0.26107326), vertexType: 1),
                          Vertex(position: float3(0.15073061, 3.763861e-08, -0.2610734), vertexType: 2),
                          Vertex(position: float3(5.476345e-09, -0.23187071, -0.4592369), vertexType: 0),
                          Vertex(position: float3(-0.22961858, -0.23187071, -0.3977107), vertexType: 1),
                          Vertex(position: float3(0.2296182, -0.23187071, -0.39771095), vertexType: 2),
                          Vertex(position: float3(0.4707132, 0.22117807, -0.47071335), vertexType: 0),
                          Vertex(position: float3(0.17229299, 0.22117807, -0.6430063), vertexType: 1),
                          Vertex(position: float3(0.6430062, 0.22117807, -0.1722935), vertexType: 2),
                          Vertex(position: float3(0.23689905, 0.12492625, -0.23689912), vertexType: 0),
                          Vertex(position: float3(0.08671107, 0.12492625, -0.32361016), vertexType: 1),
                          Vertex(position: float3(0.3236101, 0.12492625, -0.086711325), vertexType: 2),
                          Vertex(position: float3(0.236899, -0.12492617, -0.23689908), vertexType: 0),
                          Vertex(position: float3(0.08671106, -0.12492617, -0.32361013), vertexType: 1),
                          Vertex(position: float3(0.32361007, -0.12492617, -0.08671131), vertexType: 2),
                          Vertex(position: float3(0.4707132, -0.22117807, -0.47071335), vertexType: 0),
                          Vertex(position: float3(0.17229299, -0.22117807, -0.6430063), vertexType: 1),
                          Vertex(position: float3(0.6430062, -0.22117807, -0.1722935), vertexType: 2)]

    // 8 lists of indexes to vertices
   public let indices =  [[[87, 88, 80 ], [87, 80,  7 ], [87,  7, 89 ], [87, 89,  4 ], [87,  4, 77 ], [87, 77, 88],
                           [84, 85, 77 ], [84, 77,  4 ], [84,  4, 86 ], [84, 86,  1 ], [84,  1, 74 ], [84, 74, 85],
                           [24, 25, 14 ], [24, 14, 37 ], [24, 37, 26 ], [24, 26, 46 ], [24, 46, 23 ], [24, 23, 25],
                           [60, 61, 53 ], [60, 53, 76 ], [60, 76, 62 ], [60, 62, 73 ], [60, 73, 50 ], [60, 50, 61],
                           [0,  1, 86 ], [0, 86, 13 ], [0, 13,  2 ], [0,  2, 22 ], [0, 22, 95 ], [0, 95,  1],
                           [3,  4, 89 ], [3, 89, 16 ], [3, 16,  5 ], [3,  5, 13 ], [3, 13, 86 ], [3, 86,  4]],
                          [[30, 31, 20 ], [30, 20, 43 ], [30, 43, 32 ], [30, 32, 40 ], [30, 40, 17 ], [30, 17, 31],
                           [72, 73, 62 ], [72, 62, 85 ], [72, 85, 74 ], [72, 74, 94 ], [72, 94, 71 ], [72, 71, 73],
                           [93, 94, 74 ], [93, 74,  1 ], [93,  1, 95 ], [93, 95, 10 ], [93, 10, 83 ], [93, 83, 94],
                           [66, 67, 59 ], [66, 59, 82 ], [66, 82, 68 ], [66, 68, 79 ], [66, 79, 56 ], [66, 56, 67],
                           [21, 22,  2 ], [21,  2, 25 ], [21, 25, 23 ], [21, 23, 34 ], [21, 34, 11 ], [21, 11, 22],
                           [18, 19, 11 ], [18, 11, 34 ], [18, 34, 20 ], [18, 20, 31 ], [18, 31,  8 ], [18,  8, 19],
                           [33, 34, 23 ], [33, 23, 46 ], [33, 46, 35 ], [33, 35, 43 ], [33, 43, 20 ], [33, 20, 34],
                           [39, 40, 32 ], [39, 32, 55 ], [39, 55, 41 ], [39, 41, 52 ], [39, 52, 29 ], [39, 29, 40],
                           [57, 58, 47 ], [57, 47, 70 ], [57, 70, 59 ], [57, 59, 67 ], [57, 67, 44 ], [57, 44, 58],
                           [9, 10, 95 ], [9, 95, 22 ], [9, 22, 11 ], [9, 11, 19 ], [9, 19, 92 ], [9, 92, 10],
                           [81, 82, 71 ], [81, 71, 94 ], [81, 94, 83 ], [81, 83, 91 ], [81, 91, 68 ], [81, 68, 82],
                           [63, 64, 56 ], [63, 56, 79 ], [63, 79, 65 ], [63, 65, 76 ], [63, 76, 53 ], [63, 53, 64],
                           [15, 16,  8 ], [15,  8, 31 ], [15, 31, 17 ], [15, 17, 28 ], [15, 28,  5 ], [15,  5, 16],
                           [27, 28, 17 ], [27, 17, 40 ], [27, 40, 29 ], [27, 29, 37 ], [27, 37, 14 ], [27, 14, 28]],
                          [[54, 55, 44 ], [54, 44, 67 ], [54, 67, 56 ], [54, 56, 64 ], [54, 64, 41 ], [54, 41, 55],
                           [6,  7, 92 ], [6, 92, 19 ], [6, 19,  8 ], [6,  8, 16 ], [6, 16, 89 ], [6, 89,  7],
                           [36, 37, 29 ], [36, 29, 52 ], [36, 52, 38 ], [36, 38, 49 ], [36, 49, 26 ], [36, 26, 37],
                           [90, 91, 83 ], [90, 83, 10 ], [90, 10, 92 ], [90, 92,  7 ], [90,  7, 80 ], [90, 80, 91],
                           [69, 70, 50 ], [69, 50, 73 ], [69, 73, 71 ], [69, 71, 82 ], [69, 82, 59 ], [69, 59, 70],
                           [75, 76, 65 ], [75, 65, 88 ], [75, 88, 77 ], [75, 77, 85 ], [75, 85, 62 ], [75, 62, 76]],
                          [[78, 79, 68 ], [78, 68, 91 ], [78, 91, 80 ], [78, 80, 88 ], [78, 88, 65 ], [78, 65, 79],
                           [42, 43, 35 ], [42, 35, 58 ], [42, 58, 44 ], [42, 44, 55 ], [42, 55, 32 ], [42, 32, 43],
                           [48, 49, 38 ], [48, 38, 61 ], [48, 61, 50 ], [48, 50, 70 ], [48, 70, 47 ], [48, 47, 49],
                           [45, 46, 26 ], [45, 26, 49 ], [45, 49, 47 ], [45, 47, 58 ], [45, 58, 35 ], [45, 35, 46],
                           [12, 13,  5 ], [12,  5, 28 ], [12, 28, 14 ], [12, 14, 25 ], [12, 25,  2 ], [12,  2, 13],
                           [51, 52, 41 ], [51, 41, 64 ], [51, 64, 53 ], [51, 53, 61 ], [51, 61, 38 ], [51, 38, 52]]]
}

struct Mesh {
   let model = Model()
   let vertices: [Vertex]
   let submeshes: [Submesh]
   let terrains = Terrain.types
   let localUVCoords: [float2]
   let indexTables: [[UInt16]]
   var vertexCount: Int { return vertices.count }
   
   init() {
      vertices = model.vertices
      localUVCoords = model.localUVCoords
      indexTables = model.indexTables
      
      let colour = [float4(1,0,0,1),float4(0,1,0,1),float4(0,0,1,1),float4(0,1,1,1)]
      var subs = [Submesh]()
      for (i,partition) in model.indices.enumerated() {
         let submesh = Submesh(indices: Array(partition.joined()).map() {UInt16($0) },
                               uvOffset: terrains[i].uvOffset(),
                               colour: colour[i])
         subs.append(submesh)
      }
      submeshes = subs
   }
}

struct Submesh {
   let indices: [UInt16]
   let uvOffset: float2
   let colour: float4
   var indexCount: Int { return indices.count }
}

public enum Terrain: String, CaseIterable {
   public static var types: [Terrain] { return Terrain.allCases}
   
   case Lake = "Lake"
   case Plain = "Plain"
   case Desert = "Desert"
   case Hill = "Hills"

   public var distance: Int? {
      // nil means inaccessible
      switch self {
      case .Lake: return nil
      case .Plain: return 1
      case .Desert: return 2
      case .Hill: return 3
      }
   }
   
   public func uvOffset() -> float2 {
      switch self {
      case .Lake: return float2(0.0,0.5)
      case .Plain: return float2(0.25,0.5)
      case .Desert: return float2(0.5,0.5)
      case .Hill: return float2(0.75,0.5)
      }
   }
   
   public func colour() -> float3 {
      switch self {
      case .Lake: return float3(44.0/256,176.0/256,241.0/256)
      case .Plain: return float3(55.0/256,75.0/256,90.0/256)
      case .Desert: return float3(50.0/256,81.0/256,23.0/256)
      case .Hill: return float3(50.0/256,40.0/256,23.0/256)
      }
   }
}
