//
//  Shaders.metal
//  PlanetHex
//
//  Created by Don McBrien on 23/03/2020.
//  Copyright © 2020 Don McBrien. All rights reserved.
//

#include <metal_stdlib>
using namespace metal;
#import <simd/simd.h>
#import "Common.h"

struct VertexIn {
   float3 position;
   uint16_t vertexType;    // (0..2) whether created on Centre, CentreLeft or CentreRight of horizontal axis
};

struct VertexOut {
   float4 position [[position]];
   float2 uv;
   float4 colour;
};

struct Output {
   float4 position;  // as in VertexIn
   float4 colour;    // as provided in VertexUniforms
   float2 uvLocal;   // as looked up in localUVs using index
   float2 uvOffset;  // as provided in VertexUniforms
   float2 uv;        // local + global
   ushort vertexType;  // as provided in VertexIn
   ushort currentTriangle; // as provided in VertexUniforms
   ushort index;   // as calculated
};

vertex VertexOut vertex_main(constant VertexIn *vertices [[buffer(0)]],
                             constant VertexUniforms &uniforms [[buffer(1)]],
                             constant float2 *localUVs [[buffer(2)]],      // contains 7 uv pairs
                             constant ushort *indexTable [[buffer(3)]],    // contains 18 values (0...6)
                             device Output *output [[buffer(4)]],
                             uint id [[vertex_id]]) {

   float4 position = float4(vertices[id].position, 1);
   ushort index = vertices[id].vertexType * 6 + uniforms.currentTriangle;   // produces integer in  (0...17)
   float2 uvLocal = localUVs[indexTable[index]];
   
   VertexOut out;
   out.position = position;
   out.uv = uvLocal + uniforms.uvOffset;
   out.colour = uniforms.colour;
   
   output[id].position = out.position;
   output[id].colour = out.colour;
   output[id].uvLocal = uvLocal;
   output[id].uvOffset = uniforms.uvOffset;
   output[id].uv = out.uv;
   output[id].vertexType = vertices[id].vertexType;
   output[id].currentTriangle = uniforms.currentTriangle;
   output[id].index = index;
   return out;
}

fragment float4 fragment_main(VertexOut in [[stage_in]],
                              sampler textureSampler [[sampler(0)]],
                              texture2d<float> colorTexture [[texture(0)]]) {
//   float x = in.uvLocal.x * 4.0;
//   x = (x * x);
//   float y = in.uvLocal.y * 2.0;
//   y = (y * y);
//   float a = 1.0 - sqrt(x + y);
   float2 uv = in.uv; //Local + in.uvOffset;
   float3 colour = colorTexture.sample(textureSampler, uv).rgb;
   //return float4(colour,1);   // uncomment to see texture
   return in.colour;          // uncomment to see colour only
}


